@@ -14,7 +14,7 @@ MVMint64 MVM_6model_container_iscont_rw(MVMThreadContext *tc, MVMObject *cont) {
  * CodePair container configuration: container with FETCH/STORE code refs
  * ***************************************************************************/
 
- typedef struct {
+typedef struct {
     MVMObject *fetch_code;
     MVMObject *store_code;
 } CodePairContData;
@@ -122,6 +122,7 @@ static const MVMContainerSpec code_pair_spec = {
     code_pair_gc_free_data,
     code_pair_serialize,
     code_pair_deserialize,
+    NULL,
     code_pair_can_store,
     NULL, /* cas */
     NULL, /* atomic_load */
@@ -164,6 +165,268 @@ static const MVMContainerConfigurer CodePairContainerConfigurer = {
     code_pair_configure_container_spec
 };
 
+/* ***************************************************************************
+ * Value and descriptor container configuration: container is a P6opaque with
+ * an attribute designated to hold an (object) value, and other designated to
+ * be a descriptor.
+ * ***************************************************************************/
+
+/* Registered container operation callbacks. */
+typedef struct {
+    /* We cache the offsets in the P6opaque of the value and descriptor. */
+    size_t value_offset;
+    size_t descriptor_offset;
+
+    /* Callbacks. */
+    MVMObject *store;
+    MVMObject *store_unchecked;
+    MVMObject *cas;
+    MVMObject *atomic_store;
+
+    /* Retained for serializatin purposes only. */
+    MVMObject *attrs_class;
+    MVMString *value_attr;
+    MVMString *descriptor_attr;
+} MVMValueDescContainer;
+
+static void calculate_attr_offsets(MVMThreadContext *tc, MVMSTable *st, MVMValueDescContainer *data) {
+    data->value_offset = MVM_p6opaque_attr_offset(tc, st->WHAT, data->attrs_class,
+            data->value_attr) + sizeof(MVMObject);
+    data->descriptor_offset = MVM_p6opaque_attr_offset(tc, st->WHAT, data->attrs_class,
+            data->descriptor_attr) + sizeof(MVMObject);
+}
+
+static MVMObject * read_container_value(MVMObject *cont) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)STABLE(cont)->container_data;
+    return *((MVMObject **)((char*)cont + data->value_offset));
+}
+
+static MVMObject * read_container_descriptor(MVMObject *cont) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)STABLE(cont)->container_data;
+    return *((MVMObject **)((char*)cont + data->descriptor_offset));
+}
+
+static void value_desc_cont_fetch(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
+    MVMObject *value = read_container_value(cont);
+    res->o = value ? value : tc->instance->VMNull;
+}
+
+static void value_desc_cont_fetch_i(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
+    res->i64 = MVM_repr_get_int(tc, read_container_value(cont));
+}
+
+static void value_desc_cont_fetch_n(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
+    res->n64 = MVM_repr_get_num(tc, read_container_value(cont));
+}
+
+static void value_desc_cont_fetch_s(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
+    res->s = MVM_repr_get_str(tc, read_container_value(cont));
+}
+
+static void value_desc_cont_store(MVMThreadContext *tc, MVMObject *cont, MVMObject *value) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)STABLE(cont)->container_data;
+    MVMObject *code = MVM_frame_find_invokee(tc, data->store, NULL);
+    MVMCallsite *cs = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_TWO_OBJ);
+    MVM_args_setup_thunk(tc, NULL, MVM_RETURN_VOID, cs);
+    tc->cur_frame->args[0].o = cont;
+    tc->cur_frame->args[1].o = value;
+    STABLE(code)->invoke(tc, code, cs, tc->cur_frame->args);
+}
+
+static void value_desc_cont_store_i(MVMThreadContext *tc, MVMObject *cont, MVMint64 value) {
+    MVMObject *boxed;
+    MVMROOT(tc, cont, {
+        boxed = MVM_repr_box_int(tc, MVM_hll_current(tc)->int_box_type, value);
+    });
+    value_desc_cont_store(tc, cont, boxed);
+}
+
+static void value_desc_cont_store_n(MVMThreadContext *tc, MVMObject *cont, MVMnum64 value) {
+    MVMObject *boxed;
+    MVMROOT(tc, cont, {
+        boxed = MVM_repr_box_num(tc, MVM_hll_current(tc)->num_box_type, value);
+    });
+    value_desc_cont_store(tc, cont, boxed);
+}
+
+static void value_desc_cont_store_s(MVMThreadContext *tc, MVMObject *cont, MVMString *value) {
+    MVMObject *boxed;
+    MVMROOT(tc, cont, {
+        boxed = MVM_repr_box_str(tc, MVM_hll_current(tc)->str_box_type, value);
+    });
+    value_desc_cont_store(tc, cont, boxed);
+}
+
+static void value_desc_cont_store_unchecked(MVMThreadContext *tc, MVMObject *cont, MVMObject *value) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)STABLE(cont)->container_data;
+    MVMObject *code = MVM_frame_find_invokee(tc, data->store_unchecked, NULL);
+    MVMCallsite *cs = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_TWO_OBJ);
+    MVM_args_setup_thunk(tc, NULL, MVM_RETURN_VOID, cs);
+    tc->cur_frame->args[0].o = cont;
+    tc->cur_frame->args[1].o = value;
+    STABLE(code)->invoke(tc, code, cs, tc->cur_frame->args);
+}
+
+static void value_desc_cont_gc_mark_data(MVMThreadContext *tc, MVMSTable *st, MVMGCWorklist *worklist) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)st->container_data;
+    MVM_gc_worklist_add(tc, worklist, &data->store);
+    MVM_gc_worklist_add(tc, worklist, &data->store_unchecked);
+    MVM_gc_worklist_add(tc, worklist, &data->cas);
+    MVM_gc_worklist_add(tc, worklist, &data->atomic_store);
+    MVM_gc_worklist_add(tc, worklist, &data->attrs_class);
+    MVM_gc_worklist_add(tc, worklist, &data->value_attr);
+    MVM_gc_worklist_add(tc, worklist, &data->descriptor_attr);
+}
+
+static void value_desc_cont_gc_free_data(MVMThreadContext *tc, MVMSTable *st) {
+    MVM_free_null(st->container_data);
+}
+
+static void value_desc_cont_serialize(MVMThreadContext *tc, MVMSTable *st, MVMSerializationWriter *writer) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)st->container_data;
+    MVM_serialization_write_ref(tc, writer, data->store);
+    MVM_serialization_write_ref(tc, writer, data->store_unchecked);
+    MVM_serialization_write_ref(tc, writer, data->cas);
+    MVM_serialization_write_ref(tc, writer, data->atomic_store);
+    MVM_serialization_write_ref(tc, writer, data->attrs_class);
+    MVM_serialization_write_str(tc, writer, data->value_attr);
+    MVM_serialization_write_str(tc, writer, data->descriptor_attr);
+}
+
+static void value_desc_cont_deserialize(MVMThreadContext *tc, MVMSTable *st, MVMSerializationReader *reader) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)st->container_data;
+    MVM_ASSIGN_REF(tc, &(st->header), data->store, MVM_serialization_read_ref(tc, reader));
+    MVM_ASSIGN_REF(tc, &(st->header), data->store_unchecked, MVM_serialization_read_ref(tc, reader));
+    MVM_ASSIGN_REF(tc, &(st->header), data->cas, MVM_serialization_read_ref(tc, reader));
+    MVM_ASSIGN_REF(tc, &(st->header), data->atomic_store, MVM_serialization_read_ref(tc, reader));
+    MVM_ASSIGN_REF(tc, &(st->header), data->attrs_class, MVM_serialization_read_ref(tc, reader));
+    MVM_ASSIGN_REF(tc, &(st->header), data->value_attr, MVM_serialization_read_str(tc, reader));
+    MVM_ASSIGN_REF(tc, &(st->header), data->descriptor_attr, MVM_serialization_read_str(tc, reader));
+}
+
+
+static void value_desc_cont_post_deserialize(MVMThreadContext *tc, MVMSTable *st) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)st->container_data;
+    calculate_attr_offsets(tc, st, data);
+}
+
+static void value_desc_cont_spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)st->container_data;
+    switch (ins->info->opcode) {
+    case MVM_OP_decont: {
+        MVMSpeshOperand *old_operands = ins->operands;
+        ins->info = MVM_op_get_op(MVM_OP_sp_get_o);
+        ins->operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
+        ins->operands[0] = old_operands[0];
+        ins->operands[1] = old_operands[1];
+        ins->operands[2].lit_i16 = (MVMuint16)data->value_offset;
+        break;
+        }
+    default: break;
+    }
+}
+
+static MVMint32 value_desc_cont_can_store(MVMThreadContext *tc, MVMObject *cont) {
+    return !MVM_is_null(tc, read_container_descriptor(cont));
+}
+
+static void value_desc_cont_cas(MVMThreadContext *tc, MVMObject *cont,
+                              MVMObject *expected, MVMObject *value,
+                              MVMRegister *result) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)STABLE(cont)->container_data;
+    MVMObject *code = MVM_frame_find_invokee(tc, data->cas, NULL);
+    MVMCallsite *cs = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_TYPECHECK);
+    MVM_args_setup_thunk(tc, result, MVM_RETURN_OBJ, cs);
+    tc->cur_frame->args[0].o = cont;
+    tc->cur_frame->args[1].o = expected;
+    tc->cur_frame->args[2].o = value;
+    STABLE(code)->invoke(tc, code, cs, tc->cur_frame->args);
+}
+
+static MVMObject * value_desc_cont_atomic_load(MVMThreadContext *tc, MVMObject *cont) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)STABLE(cont)->container_data;
+    MVMObject *value = (MVMObject *)MVM_load(((char*)cont + data->value_offset));
+    return value ? value : tc->instance->VMNull;
+}
+
+void value_desc_cont_atomic_store(MVMThreadContext *tc, MVMObject *cont, MVMObject *value) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)STABLE(cont)->container_data;
+    MVMObject *code = MVM_frame_find_invokee(tc, data->atomic_store, NULL);
+    MVMCallsite *cs = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_TWO_OBJ);
+    MVM_args_setup_thunk(tc, NULL, MVM_RETURN_VOID, cs);
+    tc->cur_frame->args[0].o = cont;
+    tc->cur_frame->args[1].o = value;
+    STABLE(code)->invoke(tc, code, cs, tc->cur_frame->args);
+}
+
+static const MVMContainerSpec value_desc_cont_spec = {
+    "value_desc_cont",
+    value_desc_cont_fetch,
+    value_desc_cont_fetch_i,
+    value_desc_cont_fetch_n,
+    value_desc_cont_fetch_s,
+    value_desc_cont_store,
+    value_desc_cont_store_i,
+    value_desc_cont_store_n,
+    value_desc_cont_store_s,
+    value_desc_cont_store_unchecked,
+    value_desc_cont_spesh,
+    value_desc_cont_gc_mark_data,
+    value_desc_cont_gc_free_data,
+    value_desc_cont_serialize,
+    value_desc_cont_deserialize,
+    value_desc_cont_post_deserialize,
+    value_desc_cont_can_store,
+    value_desc_cont_cas,
+    value_desc_cont_atomic_load,
+    value_desc_cont_atomic_store,
+    1
+};
+
+static void value_desc_cont_set_container_spec(MVMThreadContext *tc, MVMSTable *st) {
+    if (st->container_data)
+        value_desc_cont_gc_free_data(tc, st);
+    MVMValueDescContainer *data = MVM_calloc(1, sizeof(MVMValueDescContainer));
+    st->container_data = data;
+    st->container_spec = &value_desc_cont_spec;
+}
+
+static MVMObject * grab_one_value(MVMThreadContext *tc, MVMObject *config, const char *key) {
+    MVMString *key_str;
+    MVMROOT(tc, config, {
+        key_str = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, key);
+    });
+    if (!MVM_repr_exists_key(tc, config, key_str))
+        MVM_exception_throw_adhoc(tc, "Container spec must be configured with a '%s'", key);
+    return MVM_repr_at_key_o(tc, config, key_str);
+}
+static void value_desc_cont_configure_container_spec(MVMThreadContext *tc, MVMSTable *st, MVMObject *config) {
+    MVMValueDescContainer *data = (MVMValueDescContainer *)st->container_data;
+    MVMROOT2(tc, st, config, {
+        MVMObject *value;
+        value = grab_one_value(tc, config, "store");
+        MVM_ASSIGN_REF(tc, &(st->header), data->store, value);
+        value = grab_one_value(tc, config, "store_unchecked");
+        MVM_ASSIGN_REF(tc, &(st->header), data->store_unchecked, value);
+        value = grab_one_value(tc, config, "cas");
+        MVM_ASSIGN_REF(tc, &(st->header), data->cas, value);
+        value = grab_one_value(tc, config, "atomic_store");
+        MVM_ASSIGN_REF(tc, &(st->header), data->atomic_store, value);
+        value = grab_one_value(tc, config, "attrs_class");
+        MVM_ASSIGN_REF(tc, &(st->header), data->attrs_class, value);
+        value = grab_one_value(tc, config, "value_attr");
+        MVM_ASSIGN_REF(tc, &(st->header), data->value_attr, MVM_repr_get_str(tc, value));
+        value = grab_one_value(tc, config, "descriptor_attr");
+        MVM_ASSIGN_REF(tc, &(st->header), data->descriptor_attr, MVM_repr_get_str(tc, value));
+    });
+    calculate_attr_offsets(tc, st, data);
+}
+
+static const MVMContainerConfigurer ValueDescContainerConfigurer = {
+    value_desc_cont_set_container_spec,
+    value_desc_cont_configure_container_spec
+};
+
 /* ***************************************************************************
  * Native reference container configuration
  * ***************************************************************************/
@@ -369,6 +632,7 @@ static const MVMContainerSpec native_ref_spec = {
     NULL, /* gc_free_data */
     native_ref_serialize,
     native_ref_deserialize,
+    NULL,
     native_ref_can_store,
     NULL, /* cas */
     NULL, /* atomic_load */
@@ -478,21 +742,18 @@ static const MVMContainerConfigurer NativeRefContainerConfigurer = {
 /* Adds a container configurer to the registry. */
 void MVM_6model_add_container_config(MVMThreadContext *tc, MVMString *name,
         const MVMContainerConfigurer *configurer) {
-    MVMContainerRegistry *entry;
+
+    if (!MVM_str_hash_key_is_valid(tc, name)) {
+        MVM_str_hash_key_throw_invalid(tc, name);
+    }
 
     uv_mutex_lock(&tc->instance->mutex_container_registry);
 
-    MVM_HASH_GET(tc, tc->instance->container_registry, name, entry);
+    MVMContainerRegistry *entry = MVM_str_hash_lvalue_fetch_nocheck(tc, &tc->instance->container_registry, name);
 
-    if (!entry) {
-        entry = MVM_malloc(sizeof(MVMContainerRegistry));
-        entry->name = name;
-        entry->configurer  = configurer;
-        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->name,
-            "Container configuration name");
-        MVM_HASH_BIND(tc, tc->instance->container_registry, name, entry);
-        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->hash_handle.key,
-            "Container configuration hash key");
+    if (!entry->hash_handle.key) {
+        entry->configurer      = configurer;
+        entry->hash_handle.key = name;
     }
 
     uv_mutex_unlock(&tc->instance->mutex_container_registry);
@@ -500,9 +761,15 @@ void MVM_6model_add_container_config(MVMThreadContext *tc, MVMString *name,
 
 /* Gets a container configurer from the registry. */
 const MVMContainerConfigurer * MVM_6model_get_container_config(MVMThreadContext *tc, MVMString *name) {
-    MVMContainerRegistry *entry;
-    MVM_HASH_GET(tc, tc->instance->container_registry, name, entry);
-    return entry != NULL ? entry->configurer : NULL;
+
+    if (!MVM_str_hash_key_is_valid(tc, name)) {
+        MVM_str_hash_key_throw_invalid(tc, name);
+    }
+
+    uv_mutex_lock(&tc->instance->mutex_container_registry);
+    MVMContainerRegistry *entry = MVM_str_hash_fetch_nocheck(tc, &tc->instance->container_registry, name);
+    uv_mutex_unlock(&tc->instance->mutex_container_registry);
+    return entry ? entry->configurer : NULL;
 }
 
 /* Does initial setup work of the container registry, including registering
@@ -513,6 +780,8 @@ void MVM_6model_containers_setup(MVMThreadContext *tc) {
         MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "code_pair"), &CodePairContainerConfigurer);
     MVM_6model_add_container_config(tc,
         MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "native_ref"), &NativeRefContainerConfigurer);
+    MVM_6model_add_container_config(tc,
+        MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "value_desc_cont"), &ValueDescContainerConfigurer);
 }
 
 /* ***************************************************************************
