@@ -91,6 +91,15 @@ void MVM_sc_add_all_scs_entry(MVMThreadContext *tc, MVMSerializationContextBody
     tc->instance->all_scs_next_idx++;
 }
 
+void MVM_sc_all_scs_destroy(MVMThreadContext *tc) {
+    MVM_fixed_size_free(
+        tc,
+        tc->instance->fsa,
+        tc->instance->all_scs_alloc * sizeof(MVMSerializationContextBody *),
+        tc->instance->all_scs
+    );
+}
+
 /* Given an SC, returns its unique handle. */
 MVMString * MVM_sc_get_handle(MVMThreadContext *tc, MVMSerializationContext *sc) {
     return sc->body->handle;
@@ -161,11 +170,13 @@ MVMint64 MVM_sc_find_code_idx(MVMThreadContext *tc, MVMSerializationContext *sc,
     if (REPR(obj)->ID == MVM_REPR_ID_MVMCode) {
         char *c_name = MVM_string_utf8_encode_C_string(tc, ((MVMCode *)obj)->body.name);
         char *waste[] = { c_name, NULL };
+        MVM_gc_allocate_gen2_default_clear(tc);
         MVM_exception_throw_adhoc_free(tc, waste,
             "Code ref '%s' does not exist in serialization context",
                 c_name);
     }
     else {
+        MVM_gc_allocate_gen2_default_clear(tc);
         MVM_exception_throw_adhoc(tc,
             "Code ref '<NOT A CODE OBJECT>' does not exist in serialization context");
     }
