@@ -25,8 +25,11 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
     MVMCompUnit        *cu                = static_frame_body->cu;
 
     /* Ensure the frame is fully deserialized. */
-    if (!static_frame_body->fully_deserialized)
-        MVM_bytecode_finish_frame(tc, cu, static_frame, 0);
+    if (!static_frame_body->fully_deserialized) {
+        MVMROOT(tc, static_frame, {
+            MVM_bytecode_finish_frame(tc, cu, static_frame, 0);
+        });
+    }
 
     /* If we never invoked this compilation unit before, and we have spesh
      * enabled, we might either have no spesh log or a nearly full one. This
@@ -39,11 +42,9 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
 
     /* Take compilation unit lock, to make sure we don't race to do the
      * frame preparation/verification work. */
-    MVMROOT(tc, cu, {
     MVMROOT(tc, static_frame, {
         MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
     });
-    });
 
     if (static_frame->body.instrumentation_level == 0) {
         /* Work size is number of locals/registers plus size of the maximum
@@ -52,7 +53,9 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
             (static_frame_body->num_locals + static_frame_body->cu->body.max_callsite_size);
 
         /* Validate the bytecode. */
-        MVM_validate_static_frame(tc, static_frame);
+        MVMROOT(tc, static_frame, {
+            MVM_validate_static_frame(tc, static_frame);
+        });
 
         /* Compute work area initial state that we can memcpy into place each
          * time. */
@@ -78,7 +81,7 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
          * also because then it can be ssigned into the gen2 static frame
          * without causing it to become an inter-gen root. */
         MVM_gc_allocate_gen2_default_set(tc);
-        MVM_ASSIGN_REF(tc, &(static_frame->common.header), static_frame_body->spesh,
+        MVM_ASSIGN_REF(tc, &(static_frame->common.header), static_frame_body->spesh, /* no GC error */
             MVM_repr_alloc_init(tc, tc->instance->StaticFrameSpesh));
         MVM_gc_allocate_gen2_default_clear(tc);
 
@@ -96,7 +99,7 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
  * profiling. */
 static void instrumentation_level_barrier(MVMThreadContext *tc, MVMStaticFrame *static_frame) {
     MVMCompUnit *cu = static_frame->body.cu;
-    MVMROOT2(tc, static_frame, cu, {
+    MVMROOT(tc, static_frame, {
         /* Obtain mutex, so we don't end up with instrumentation races. */
         MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
 
@@ -262,7 +265,7 @@ static MVMFrame * allocate_frame(MVMThreadContext *tc, MVMStaticFrame *static_fr
 
     if (heap) {
         /* Allocate frame on the heap. We know it's already zeroed. */
-        MVMROOT(tc, static_frame, {
+        MVMROOT2(tc, static_frame, spesh_cand, {
             if (tc->cur_frame)
                 MVM_frame_force_to_heap(tc, tc->cur_frame);
             frame = MVM_gc_allocate_frame(tc);
@@ -278,7 +281,8 @@ static MVMFrame * allocate_frame(MVMThreadContext *tc, MVMStaticFrame *static_fr
 
         /* Ensure collectable header flags and owner are zeroed, which means we'll
          * never try to mark or root the frame. */
-        frame->header.flags = 0;
+        frame->header.flags1 = 0;
+        frame->header.flags2 = 0;
         frame->header.owner = 0;
 
         /* Current arguments callsite must be NULL as it's used in GC. Extra must
@@ -290,11 +294,11 @@ static MVMFrame * allocate_frame(MVMThreadContext *tc, MVMStaticFrame *static_fr
 
     /* Allocate space for lexicals and work area. */
     static_frame_body = &(static_frame->body);
-    env_size = spesh_cand ? spesh_cand->env_size : static_frame_body->env_size;
+    env_size = spesh_cand ? spesh_cand->body.env_size : static_frame_body->env_size;
 
-    jitcode = spesh_cand ? spesh_cand->jitcode : NULL;
+    jitcode = spesh_cand ? spesh_cand->body.jitcode : NULL;
     num_locals = jitcode && jitcode->local_types ? jitcode->num_locals :
-        (spesh_cand ? spesh_cand->num_locals : static_frame_body->num_locals);
+        (spesh_cand ? spesh_cand->body.num_locals : static_frame_body->num_locals);
     if (env_size) {
         frame->env = MVM_fixed_size_alloc_zeroed(tc, tc->instance->fsa, env_size);
         frame->allocd_env = env_size;
@@ -303,7 +307,7 @@ static MVMFrame * allocate_frame(MVMThreadContext *tc, MVMStaticFrame *static_fr
         frame->env = NULL;
         frame->allocd_env = 0;
     }
-    work_size = spesh_cand ? spesh_cand->work_size : static_frame_body->work_size;
+    work_size = spesh_cand ? spesh_cand->body.work_size : static_frame_body->work_size;
     if (work_size) {
         if (spesh_cand) {
             /* Allocate zeroed memory. Spesh makes sure we have VMNull setup in
@@ -345,18 +349,18 @@ static MVMFrame * allocate_heap_frame(MVMThreadContext *tc, MVMStaticFrame *stat
     MVMStaticFrameBody *static_frame_body;
 
     /* Allocate the frame. */
-    MVMROOT(tc, static_frame, {
+    MVMROOT2(tc, static_frame, spesh_cand, {
         frame = MVM_gc_allocate_frame(tc);
     });
 
     /* Allocate space for lexicals and work area. */
     static_frame_body = &(static_frame->body);
-    env_size = spesh_cand ? spesh_cand->env_size : static_frame_body->env_size;
+    env_size = spesh_cand ? spesh_cand->body.env_size : static_frame_body->env_size;
     if (env_size) {
         frame->env = MVM_fixed_size_alloc_zeroed(tc, tc->instance->fsa, env_size);
         frame->allocd_env = env_size;
     }
-    work_size = spesh_cand ? spesh_cand->work_size : static_frame_body->work_size;
+    work_size = spesh_cand ? spesh_cand->body.work_size : static_frame_body->work_size;
     if (work_size) {
         if (spesh_cand) {
             /* We make sure to initialize anything that can be read before
@@ -374,7 +378,7 @@ static MVMFrame * allocate_heap_frame(MVMThreadContext *tc, MVMStaticFrame *stat
 
         /* Calculate args buffer position. */
         frame->args = frame->work + (spesh_cand
-            ? spesh_cand->num_locals
+            ? spesh_cand->body.num_locals
             : static_frame_body->num_locals);
     }
 
@@ -471,7 +475,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
         }
         else if (static_frame->body.outer) {
             /* Auto-close, and cache it in the static frame. */
-            MVMROOT2(tc, static_frame, code_ref, {
+            MVMROOT3(tc, static_frame, code_ref, static_code, {
                 MVM_frame_force_to_heap(tc, tc->cur_frame);
                 outer = autoclose(tc, static_frame->body.outer);
                 MVM_ASSIGN_REF(tc, &(static_code->common.header),
@@ -505,7 +509,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
     if (spesh_cand >= 0) {
         MVMSpeshCandidate *chosen_cand = spesh->body.spesh_candidates[spesh_cand];
         if (static_frame->body.allocate_on_heap) {
-            MVMROOT3(tc, static_frame, code_ref, outer, {
+            MVMROOT4(tc, static_frame, code_ref, outer, chosen_cand, {
                 frame = allocate_frame(tc, static_frame, chosen_cand, 1);
             });
         }
@@ -513,15 +517,15 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
             frame = allocate_frame(tc, static_frame, chosen_cand, 0);
             frame->spesh_correlation_id = 0;
         }
-        if (chosen_cand->jitcode) {
-            chosen_bytecode = chosen_cand->jitcode->bytecode;
-            frame->jit_entry_label = chosen_cand->jitcode->labels[0];
+        if (chosen_cand->body.jitcode) {
+            chosen_bytecode = chosen_cand->body.jitcode->bytecode;
+            frame->jit_entry_label = chosen_cand->body.jitcode->labels[0];
         }
         else {
-            chosen_bytecode = chosen_cand->bytecode;
+            chosen_bytecode = chosen_cand->body.bytecode;
         }
-        frame->effective_spesh_slots = chosen_cand->spesh_slots;
-        frame->spesh_cand = chosen_cand;
+        frame->effective_spesh_slots = chosen_cand->body.spesh_slots;
+        MVM_ASSIGN_REF(tc, &(frame->header), frame->spesh_cand, chosen_cand);
     }
     else {
         MVMint32 on_heap = static_frame->body.allocate_on_heap;
@@ -645,6 +649,7 @@ MVMFrame * MVM_frame_move_to_heap(MVMThreadContext *tc, MVMFrame *frame) {
     MVMROOT4(tc, new_cur_frame, update_caller, cur_to_promote, result, {
         while (cur_to_promote) {
             /* Allocate a heap frame. */
+            /* frame is safe from the GC as we wouldn't be here if it wasn't on the stack */
             MVMFrame *promoted = MVM_gc_allocate_frame(tc);
 
             /* Bump heap promotion counter, to encourage allocating this kind
@@ -749,9 +754,10 @@ MVMFrame * MVM_frame_debugserver_move_to_heap(MVMThreadContext *tc, MVMThreadCon
     MVMFrame *update_caller = NULL;
     MVMFrame *result = NULL;
     MVM_CHECK_CALLER_CHAIN(tc, cur_to_promote);
-    MVMROOT3(tc, new_cur_frame, update_caller, result, {
+    MVMROOT4(tc, new_cur_frame, update_caller, cur_to_promote, result, {
         while (cur_to_promote) {
             /* Allocate a heap frame. */
+            /* frame is safe from the GC as we wouldn't be here if it wasn't on the stack */
             MVMFrame *promoted = MVM_gc_allocate_frame(tc);
 
             /* Bump heap promotion counter, to encourage allocating this kind
@@ -917,7 +923,7 @@ static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
 
        if (tc->jit_return_address != NULL) {
             /* on a JIT frame, exit to interpreter afterwards */
-            MVMJitCode *jitcode = returner->spesh_cand->jitcode;
+            MVMJitCode *jitcode = returner->spesh_cand->body.jitcode;
             MVM_jit_code_set_current_position(tc, jitcode, returner, jitcode->exit_label);
             /* given that we might throw in the special-return, act as if we've
              * left the current frame (which is true) */
@@ -1070,6 +1076,9 @@ static void continue_unwind(MVMThreadContext *tc, void *sr_data) {
     MVM_free(sr_data);
     MVM_frame_unwind_to(tc, frame, abs_addr, rel_addr, NULL, jit_return_label);
 }
+static void free_unwind_data(MVMThreadContext *tc, void *sr_data) {
+    MVM_free(sr_data);
+}
 void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_addr,
                          MVMuint32 rel_addr, MVMObject *return_value, void *jit_return_label) {
     while (tc->cur_frame != frame) {
@@ -1084,6 +1093,9 @@ void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_ad
             MVMObject    *handler;
             MVMCallsite *two_args_callsite;
 
+            if (return_value)
+                MVM_exception_throw_adhoc(tc, "return_value + exit_handler case NYI");
+
             /* Force the frame onto the heap, since we'll reference it from the
              * unwind data. */
             MVMROOT3(tc, frame, cur_frame, return_value, {
@@ -1108,9 +1120,7 @@ void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_ad
                 ud->abs_addr = abs_addr;
                 ud->rel_addr = rel_addr;
                 ud->jit_return_label = jit_return_label;
-                if (return_value)
-                    MVM_exception_throw_adhoc(tc, "return_value + exit_handler case NYI");
-                MVM_frame_special_return(tc, cur_frame, continue_unwind, NULL, ud,
+                MVM_frame_special_return(tc, cur_frame, continue_unwind, free_unwind_data, ud,
                     mark_unwind_data);
             }
             cur_frame->flags |= MVM_FRAME_FLAG_EXIT_HAND_RUN;
@@ -1133,7 +1143,7 @@ void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_ad
         *tc->interp_cur_op = *tc->interp_bytecode_start + rel_addr;
 
     if (jit_return_label) {
-        MVM_jit_code_set_current_position(tc, tc->cur_frame->spesh_cand->jitcode, tc->cur_frame, jit_return_label);
+        MVM_jit_code_set_current_position(tc, tc->cur_frame->spesh_cand->body.jitcode, tc->cur_frame, jit_return_label);
     }
 
     if (return_value)
@@ -1252,9 +1262,9 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
     else if (f->spesh_cand) {
         MVMuint32 i;
         flags = NULL;
-        for (i = 0; i < f->spesh_cand->num_inlines; i++) {
-            MVMStaticFrame *isf = f->spesh_cand->inlines[i].sf;
-            effective_idx = idx - f->spesh_cand->inlines[i].lexicals_start;
+        for (i = 0; i < f->spesh_cand->body.num_inlines; i++) {
+            MVMStaticFrame *isf = f->spesh_cand->body.inlines[i].sf;
+            effective_idx = idx - f->spesh_cand->body.inlines[i].lexicals_start;
             if (effective_idx < isf->body.num_lexicals) {
                 flags        = isf->body.static_env_flags;
                 static_env   = isf->body.static_env;
@@ -1283,7 +1293,7 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
             if (sc == NULL)
                 MVM_exception_throw_adhoc(tc,
                     "SC not yet resolved; lookup failed");
-            MVMROOT(tc, f, {
+            MVMROOT2(tc, f, effective_sf, {
                 resolved = MVM_sc_get_object(tc, sc, objid);
             });
             MVM_ASSIGN_REF(tc, &(effective_sf->common.header),
@@ -1318,16 +1328,13 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
 MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type) {
     MVMFrame *cur_frame = tc->cur_frame;
     while (cur_frame != NULL) {
-        MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
-        if (lexical_names) {
-            /* Indexes were formerly stored off-by-one to avoid semi-predicate issue. */
-            MVMLexicalRegistry *entry;
-            MVM_HASH_GET(tc, lexical_names, name, entry)
-            if (entry) {
-                if (MVM_LIKELY(cur_frame->static_info->body.lexical_types[entry->value] == type)) {
-                    MVMRegister *result = &cur_frame->env[entry->value];
+        if (cur_frame->static_info->body.num_lexicals) {
+            MVMuint32 idx = MVM_get_lexical_by_name(tc, cur_frame->static_info, name);
+            if (idx != MVM_INDEX_HASH_NOT_FOUND) {
+                if (MVM_LIKELY(cur_frame->static_info->body.lexical_types[idx] == type)) {
+                    MVMRegister *result = &cur_frame->env[idx];
                     if (type == MVM_reg_obj && !result->o)
-                        MVM_frame_vivify_lexical(tc, cur_frame, entry->value);
+                        MVM_frame_vivify_lexical(tc, cur_frame, idx);
                     return result;
                 }
                 else {
@@ -1355,18 +1362,16 @@ MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *na
 MVM_PUBLIC void MVM_frame_bind_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type, MVMRegister value) {
     MVMFrame *cur_frame = tc->cur_frame;
     while (cur_frame != NULL) {
-        MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
-        if (lexical_names) {
-            MVMLexicalRegistry *entry;
-            MVM_HASH_GET(tc, lexical_names, name, entry)
-            if (entry) {
-                if (cur_frame->static_info->body.lexical_types[entry->value] == type) {
+        if (cur_frame->static_info->body.num_lexicals) {
+            MVMuint32 idx = MVM_get_lexical_by_name(tc, cur_frame->static_info, name);
+            if (idx != MVM_INDEX_HASH_NOT_FOUND) {
+                if (cur_frame->static_info->body.lexical_types[idx] == type) {
                     if (type == MVM_reg_obj || type == MVM_reg_str) {
                         MVM_ASSIGN_REF(tc, &(cur_frame->header),
-                            cur_frame->env[entry->value].o, value.o);
+                            cur_frame->env[idx].o, value.o);
                     }
                     else {
-                        cur_frame->env[entry->value] = value;
+                        cur_frame->env[idx] = value;
                     }
                     return;
                 }
@@ -1391,7 +1396,10 @@ MVM_PUBLIC void MVM_frame_bind_lexical_by_name(MVMThreadContext *tc, MVMString *
 
 /* Finds a lexical in the outer frame, throwing if it's not there. */
 MVMObject * MVM_frame_find_lexical_by_name_outer(MVMThreadContext *tc, MVMString *name) {
-    MVMRegister *r = MVM_frame_find_lexical_by_name_rel(tc, name, tc->cur_frame->outer);
+    MVMRegister *r;
+    MVMROOT(tc, name, {
+        r = MVM_frame_find_lexical_by_name_rel(tc, name, tc->cur_frame->outer);
+    });
     if (MVM_LIKELY(r != NULL))
         return r->o;
     else {
@@ -1406,16 +1414,13 @@ MVMObject * MVM_frame_find_lexical_by_name_outer(MVMThreadContext *tc, MVMString
  * the specified frame. Only works if it's an object lexical.  */
 MVMRegister * MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString *name, MVMFrame *cur_frame) {
     while (cur_frame != NULL) {
-        MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
-        if (lexical_names) {
-            /* Indexes were formerly stored off-by-one to avoid semi-predicate issue. */
-            MVMLexicalRegistry *entry;
-            MVM_HASH_GET(tc, lexical_names, name, entry)
-            if (entry) {
-                if (cur_frame->static_info->body.lexical_types[entry->value] == MVM_reg_obj) {
-                    MVMRegister *result = &cur_frame->env[entry->value];
+        if (cur_frame->static_info->body.num_lexicals) {
+            MVMuint32 idx = MVM_get_lexical_by_name(tc, cur_frame->static_info, name);
+            if (idx != MVM_INDEX_HASH_NOT_FOUND) {
+                if (cur_frame->static_info->body.lexical_types[idx] == MVM_reg_obj) {
+                    MVMRegister *result = &cur_frame->env[idx];
                     if (!result->o)
-                        MVM_frame_vivify_lexical(tc, cur_frame, entry->value);
+                        MVM_frame_vivify_lexical(tc, cur_frame, idx);
                     return result;
                 }
                 else {
@@ -1436,12 +1441,14 @@ MVMRegister * MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString
  * that is done depends on the frame walker setup. */
 MVMRegister * MVM_frame_lexical_lookup_using_frame_walker(MVMThreadContext *tc,
         MVMSpeshFrameWalker *fw, MVMString *name) {
+    MVM_gc_root_temp_push(tc, (MVMCollectable **)&(name));
     while (MVM_spesh_frame_walker_next(tc, fw)) {
         MVMRegister *found;
         MVMuint16 found_kind;
         if (MVM_spesh_frame_walker_get_lex(tc, fw, name, &found, &found_kind, 1, NULL)) {
             MVM_spesh_frame_walker_cleanup(tc, fw);
             if (found_kind == MVM_reg_obj) {
+                MVM_gc_root_temp_pop(tc);
                 return found;
             }
             else {
@@ -1454,6 +1461,7 @@ MVMRegister * MVM_frame_lexical_lookup_using_frame_walker(MVMThreadContext *tc,
         }
     }
     MVM_spesh_frame_walker_cleanup(tc, fw);
+    MVM_gc_root_temp_pop(tc);
     return NULL;
 }
 
@@ -1522,6 +1530,7 @@ MVMRegister * MVM_frame_find_dynamic_using_frame_walker(MVMThreadContext *tc,
     /* Traverse with the frame walker. */
     MVM_gc_root_temp_push(tc, (MVMCollectable **)&initial_frame);
     MVM_gc_root_temp_push(tc, (MVMCollectable **)&last_real_frame);
+    MVM_gc_root_temp_push(tc, (MVMCollectable **)&name);
     while (MVM_spesh_frame_walker_next(tc, fw)) {
         MVMRegister *result;
 
@@ -1546,7 +1555,7 @@ MVMRegister * MVM_frame_find_dynamic_using_frame_walker(MVMThreadContext *tc,
                         tc->instance->dynvar_log_lasttime = uv_hrtime();
                     }
                     *found_frame = last_real_frame;
-                    MVM_gc_root_temp_pop_n(tc, 2);
+                    MVM_gc_root_temp_pop_n(tc, 3);
                     MVM_spesh_frame_walker_cleanup(tc, fw);
                     return result;
                 }
@@ -1574,14 +1583,14 @@ MVMRegister * MVM_frame_find_dynamic_using_frame_walker(MVMThreadContext *tc,
                 MVM_free(c_name);
                 tc->instance->dynvar_log_lasttime = uv_hrtime();
             }
-            MVM_gc_root_temp_pop_n(tc, 2);
+            MVM_gc_root_temp_pop_n(tc, 3);
             MVM_spesh_frame_walker_cleanup(tc, fw);
             return result;
         }
     }
 
     /* Not found. */
-    MVM_gc_root_temp_pop_n(tc, 2);
+    MVM_gc_root_temp_pop_n(tc, 3);
     MVM_spesh_frame_walker_cleanup(tc, fw);
     if (dlog) {
         fprintf(dlog, "N %s %d %d %d %d %"PRIu64" %"PRIu64" %"PRIu64"\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
@@ -1663,7 +1672,10 @@ MVMObject * MVM_frame_getdynlex(MVMThreadContext *tc, MVMString *name, MVMFrame
 void MVM_frame_binddynlex(MVMThreadContext *tc, MVMString *name, MVMObject *value, MVMFrame *cur_frame) {
     MVMuint16 type;
     MVMFrame *found_frame;
-    MVMRegister *lex_reg = MVM_frame_find_contextual_by_name(tc, name, &type, cur_frame, 0, &found_frame);
+    MVMRegister *lex_reg;
+    MVMROOT2(tc, name, value, {
+        lex_reg = MVM_frame_find_contextual_by_name(tc, name, &type, cur_frame, 0, &found_frame);
+    });
     if (!lex_reg) {
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
         char *waste[] = { c_name, NULL };
@@ -1694,12 +1706,10 @@ void MVM_frame_binddynlex(MVMThreadContext *tc, MVMString *name, MVMObject *valu
 /* Returns the storage unit for the lexical in the specified frame. Does not
  * try to vivify anything - gets exactly what is there. */
 MVMRegister * MVM_frame_lexical(MVMThreadContext *tc, MVMFrame *f, MVMString *name) {
-    MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
-    if (MVM_LIKELY(lexical_names != NULL)) {
-        MVMLexicalRegistry *entry;
-        MVM_HASH_GET(tc, lexical_names, name, entry)
-        if (entry)
-            return &f->env[entry->value];
+    if (MVM_LIKELY(f->static_info->body.num_lexicals != 0)) {
+        MVMuint32 idx = MVM_get_lexical_by_name(tc, f->static_info, name);
+        if (idx != MVM_INDEX_HASH_NOT_FOUND)
+            return &f->env[idx];
     }
     {
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
@@ -1711,14 +1721,12 @@ MVMRegister * MVM_frame_lexical(MVMThreadContext *tc, MVMFrame *f, MVMString *na
 
 /* Returns the storage unit for the lexical in the specified frame. */
 MVMRegister * MVM_frame_try_get_lexical(MVMThreadContext *tc, MVMFrame *f, MVMString *name, MVMuint16 type) {
-    MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
-    if (lexical_names) {
-        MVMLexicalRegistry *entry;
-        MVM_HASH_GET(tc, lexical_names, name, entry)
-        if (entry && f->static_info->body.lexical_types[entry->value] == type) {
-            MVMRegister *result = &f->env[entry->value];
+    if (f->static_info->body.num_lexicals) {
+        MVMuint32 idx = MVM_get_lexical_by_name(tc, f->static_info, name);
+        if (idx != MVM_INDEX_HASH_NOT_FOUND && f->static_info->body.lexical_types[idx] == type) {
+            MVMRegister *result = &f->env[idx];
             if (type == MVM_reg_obj && !result->o)
-                MVM_frame_vivify_lexical(tc, f, entry->value);
+                MVM_frame_vivify_lexical(tc, f, idx);
             return result;
         }
     }
@@ -1759,13 +1767,11 @@ MVMuint16 MVM_frame_translate_to_primspec(MVMThreadContext *tc, MVMuint16 kind)
 
 /* Returns the primitive type specification for a lexical. */
 MVMuint16 MVM_frame_lexical_primspec(MVMThreadContext *tc, MVMFrame *f, MVMString *name) {
-    MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
-    if (lexical_names) {
-        MVMLexicalRegistry *entry;
-        MVM_HASH_GET(tc, lexical_names, name, entry)
-        if (entry)
+    if (f->static_info->body.num_lexicals) {
+        MVMuint32 idx = MVM_get_lexical_by_name(tc, f->static_info, name);
+        if (idx != MVM_INDEX_HASH_NOT_FOUND)
             return MVM_frame_translate_to_primspec(tc,
-                    f->static_info->body.lexical_types[entry->value]);
+                    f->static_info->body.lexical_types[idx]);
     }
     {
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
