@@ -52,7 +52,7 @@ MVMThreadContext * MVM_tc_create(MVMThreadContext *parent, MVMInstance *instance
             }
             MVM_free(tc->temp_bigints[i]);
             MVM_tc_destroy(tc);
-            MVM_exception_throw_adhoc(tc, "Error creating a temporary big integer: %s", mp_error_to_string(err));
+            MVM_exception_throw_adhoc(parent, "Error creating a temporary big integer: %s", mp_error_to_string(err));
         }
     }
 
@@ -82,14 +82,18 @@ MVMThreadContext * MVM_tc_create(MVMThreadContext *parent, MVMInstance *instance
 void MVM_tc_destroy(MVMThreadContext *tc) {
     MVMint32 i;
 
-    /* Free the native callback cache. Needs the fixed size allocator. */
-    if (tc->native_callback_cache) {
-        MVM_str_hash_demolish(tc, tc->native_callback_cache);
-        MVM_fixed_size_free(tc, tc->instance->fsa,
-                            sizeof(MVMStrHashTable),
-                            tc->native_callback_cache);
+    /* If an exception handler calls nqp::exit, we don't unwind the stack and
+     * exception handlers aren't cleaned up yet */
+    while (tc->active_handlers) {
+        MVMActiveHandler *ah = tc->active_handlers;
+        tc->active_handlers = ah->next_handler;
+        MVM_free(ah);
     }
 
+    /* Free the native callback cache. Needs the fixed size allocator. */
+    /* (currently not. but might if MVMStrHash moves internally to the FSA.) */
+    MVM_str_hash_demolish(tc, &tc->native_callback_cache);
+
     /* Free specialization state. */
     MVM_spesh_sim_stack_destroy(tc, tc->spesh_sim_stack);
 
