@@ -1,7 +1,6 @@
 #include "platform/memmem.h"
 #include "moar.h"
 #include "platform/io.h"
-#include <platform/threads.h>
 #include "platform/random.h"
 #include "platform/time.h"
 #include "platform/mmap.h"
@@ -87,6 +86,19 @@ MVM_STATIC_INLINE MVMuint64 ptr_hash_64_to_64(MVMuint64 u) {
     return (MVMuint64)u;
 }
 
+#ifdef MVM_THREAD_LOCAL
+MVM_THREAD_LOCAL MVMThreadContext *MVM_running_threads_context;
+#else
+uv_key_t MVM_running_threads_context_key;
+
+static void
+make_uv_key() {
+    int result = uv_key_create(&MVM_running_threads_context_key);
+    if (result)
+        MVM_panic(1, "uv_key_create failed with code %u", result);
+}
+#endif
+
 /* Create a new instance of the VM. */
 MVMInstance * MVM_vm_create_instance(void) {
     MVMInstance *instance;
@@ -98,6 +110,11 @@ MVMInstance * MVM_vm_create_instance(void) {
     char *dynvar_log;
     int init_stat;
 
+#ifndef MVM_THREAD_LOCAL
+    static uv_once_t key_once = UV_ONCE_INIT;
+    uv_once(&key_once, make_uv_key);
+#endif
+
     /* Set up instance data structure. */
     instance = MVM_calloc(1, sizeof(MVMInstance));
 
@@ -211,6 +228,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     instance->threads->body.tc = instance->main_thread;
     instance->threads->body.native_thread_id = MVM_platform_thread_id();
     instance->threads->body.thread_id = instance->main_thread->thread_id;
+    MVM_set_running_threads_context(instance->main_thread);
     init_mutex(instance->mutex_threads, "threads list");
 
     /* Create compiler registry */
@@ -626,10 +644,15 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     /* Run the normal GC one more time to actually collect the spesh thread */
     MVM_gc_enter_from_allocator(instance->main_thread);
 
+    MVM_profile_instrumented_free_data(instance->main_thread);
+
     /* Run the GC global destruction phase. After this,
      * no 6model object pointers should be accessed. */
     MVM_gc_global_destruction(instance->main_thread);
 
+    MVM_ptr_hash_demolish(instance->main_thread, &instance->object_ids);
+    MVM_sc_all_scs_destroy(instance->main_thread);
+
     /* Cleanup REPR registry */
     uv_mutex_destroy(&instance->mutex_repr_registry);
     MVM_index_hash_demolish(instance->main_thread, &instance->repr_hash);
